from django.test import TestCase

from ...constants import NOT_REQUIRED, REQUIRED
from ...metadata_rules import Logic, RuleLogicError


class MetadataRulesTests(TestCase):
    def test_logic(self):
        logic = Logic(
            predicate=lambda x: True if x else False,
            consequence=REQUIRED,
            alternative=NOT_REQUIRED,
        )
        self.assertTrue(logic.predicate(1) is True)
        self.assertTrue(logic.consequence == REQUIRED)
        self.assertTrue(logic.alternative == NOT_REQUIRED)
        self.assertTrue(repr(logic))

    def test_logic_invalid_consequence(self):
        self.assertRaises(
            RuleLogicError,
            Logic,
            predicate=lambda x: False if x else True,
            consequence="blah",
            alternative=NOT_REQUIRED,
        )

    def test_logic_invalid_alternative(self):
        self.assertRaises(
            RuleLogicError,
            Logic,
            predicate=lambda x: False if x else True,
            consequence=NOT_REQUIRED,
            alternative="blah",
        )

    def test_logic_assert_predicate_is_callable(self):
        self.assertRaises(
            RuleLogicError,
            Logic,
            predicate="erik",
            consequence=REQUIRED,
            alternative=NOT_REQUIRED,
        )



# Generated by Django 4.2.1 on 2023-07-07 19:32

from django.db import migrations
import edc_metadata.managers
import edc_sites.model_mixins


class Migration(migrations.Migration):
    dependencies = [
        ("edc_metadata", "0020_alter_crfmetadata_options_and_more"),
    ]

    operations = [
        migrations.AlterModelManagers(
            name="crfmetadata",
            managers=[
                ("objects", edc_metadata.managers.CrfMetadataManager()),
                ("on_site", edc_sites.model_mixins.CurrentSiteManager()),
            ],
        ),
        migrations.AlterModelManagers(
            name="requisitionmetadata",
            managers=[
                ("objects", edc_metadata.managers.RequisitionMetadataManager()),
                ("on_site", edc_sites.model_mixins.CurrentSiteManager()),
            ],
        ),
    ]



# Generated by Django 4.0.4 on 2022-05-25 16:10

from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ("edc_metadata", "0018_auto_20200513_0023"),
    ]

    operations = [
        migrations.AlterField(
            model_name="crfmetadata",
            name="entry_status",
            field=models.CharField(
                choices=[
                    ("REQUIRED", "New"),
                    ("KEYED", "Keyed"),
                    ("missed", "Missed"),
                    ("NOT_REQUIRED", "Not required"),
                ],
                db_index=True,
                default="REQUIRED",
                max_length=25,
            ),
        ),
        migrations.AlterField(
            model_name="requisitionmetadata",
            name="entry_status",
            field=models.CharField(
                choices=[
                    ("REQUIRED", "New"),
                    ("KEYED", "Keyed"),
                    ("missed", "Missed"),
                    ("NOT_REQUIRED", "Not required"),
                ],
                db_index=True,
                default="REQUIRED",
                max_length=25,
            ),
        ),
    ]



from __future__ import annotations

from typing import TYPE_CHECKING

from django.apps import apps as django_apps
from django.core.exceptions import ObjectDoesNotExist

from .constants import REQUIRED
from .metadata import CrfMetadataGetter, RequisitionMetadataGetter

if TYPE_CHECKING:
    from edc_appointment.models import Appointment
    from edc_visit_schedule import Crf, Requisition, Visit

    from .metadata import MetadataGetter
    from .models import CrfMetadata, RequisitionMetadata


class NextFormGetter:
    crf_metadata_getter_cls = CrfMetadataGetter
    requisition_metadata_getter_cls = RequisitionMetadataGetter

    def __init__(
        self,
        model_obj=None,
        appointment: Appointment = None,
        model: str = None,
        panel_name: str = None,
    ):
        self._getter = None
        self._next_metadata_obj = None
        self._model_obj = model_obj
        self._next_form = None
        self._next_panel = None
        self._panel_name = panel_name

        self.model = model or model_obj._meta.label_lower
        self.appointment = appointment or model_obj.related_visit.appointment
        self.visit: Visit = self.appointment.related_visit.visit

    @property
    def next_form(self) -> Crf | Requisition:
        """Returns the next required form based on the metadata.

        A form is a Crf or Requisition object from edc_visit_schedule.
        """
        if not self._next_form:
            next_model = getattr(self.next_metadata_obj, "model", None)
            self._next_form = self.visit.get_requisition(
                next_model, panel_name=self.next_panel
            ) or self.visit.get_crf(next_model)
        return self._next_form

    @property
    def model_obj(self):
        """Returns the model instance of the current CRF or
        Requisition.
        """
        if not self._model_obj:
            model_cls = django_apps.get_model(self.model)
            try:
                self._model_obj = model_cls.objects.get(
                    **{
                        f"{model_cls.related_visit_model_attr()}__appointment": (
                            self.appointment
                        )
                    }
                )
            except ObjectDoesNotExist:
                pass
        return self._model_obj

    @property
    def metadata_getter(self) -> MetadataGetter | RequisitionMetadataGetter:
        """Returns a metadata_getter instance."""
        if not self._getter:
            if self.panel_name:
                self._getter = self.requisition_metadata_getter_cls(
                    appointment=self.appointment
                )
            else:
                self._getter = self.crf_metadata_getter_cls(appointment=self.appointment)
        return self._getter

    @property
    def next_metadata_obj(self) -> CrfMetadata | RequisitionMetadata:
        """Returns the "next" metadata model instance or None."""
        if not self._next_metadata_obj:
            show_order = getattr(self.crf_or_requisition, "show_order", None)
            self._next_metadata_obj = self.metadata_getter.next_object(
                show_order=show_order, entry_status=REQUIRED
            )
        return self._next_metadata_obj

    @property
    def next_panel(self) -> str | None:
        if not self._next_panel:
            if self.next_metadata_obj:
                try:
                    self._next_panel = self.next_metadata_obj.panel_name
                except AttributeError:
                    pass
        return self._next_panel

    @property
    def panel_name(self) -> str | None:
        """Returns a panel_name or None."""
        if not self._panel_name:
            if self.model_obj:
                try:
                    self._panel_name = self.model_obj.panel.name
                except AttributeError:
                    self._panel_name = None
        return self._panel_name

    @property
    def crf_or_requisition(self) -> Crf | Requisition:
        """Returns a CRF or Requisition object from
        the visit schedule's visit.
        """
        crf = None
        requisition = None
        if self.panel_name:
            requisition = self.visit.get_requisition(
                model=self.model, panel_name=self.panel_name
            )
        else:
            crf = self.visit.get_crf(model=self.model)
        return crf or requisition



import sys

from django.apps import apps as django_apps
from django.core.management.color import color_style

style = color_style()


class UpdateMetadataError(Exception):
    pass


class UpdateMetadataOnScheduleChange:
    def __init__(self, field=None, old_name=None, new_name=None, dry_run=None):
        self.fieldname = field
        self.new_value = new_name
        self.old_value = old_name
        self.dry_run = dry_run
        self.update()

    def update(self):
        sys.stdout.write(style.SUCCESS("\n\nUpdate metadata.\n"))
        sys.stdout.write(style.SUCCESS(f"Field is '{self.fieldname}'.\n"))
        sys.stdout.write(
            style.SUCCESS(f"Old value='{self.old_value}', New value='{self.new_value}'.\n")
        )
        if self.dry_run:
            sys.stdout.write(
                style.WARNING("This is a dry run. No records will be updated. \n")
            )
            sys.stdout.write(
                "These models need to updated with the new "
                f"value for field '{self.fieldname}'.\n"
                f"Old value='{self.old_value}', New value='{self.new_value}'.\n"
            )
            for name, model_cls in self.models.items():
                count = model_cls.objects.filter(**{self.fieldname: self.old_value}).count()
                sys.stdout.write(f"{model_cls._meta.label_lower}. {count} records found.\n")
            sys.stdout.write(
                style.ERROR(
                    "No records have been updated. \n" "Set --dry-run=False to update.\n"
                )
            )
        else:
            sys.stdout.write(style.SUCCESS("Updating... \n"))
            for name, model_cls in self.models.items():
                sys.stdout.write(f"Updating {name} ...\r")
                updated = model_cls.objects.filter(**{self.fieldname: self.old_value}).update(
                    **{self.fieldname: self.new_value}
                )
                sys.stdout.write(f"Updated {name}. {updated} records.     \n")
            sys.stdout.write(style.SUCCESS("Done. \n"))

    @property
    def models(self):
        """Returns a dictionary of {name: model_cls} for models
        that have the field.
        """
        models = {}
        for model in django_apps.get_models():
            if [f.name for f in model._meta.get_fields() if f.name == self.fieldname]:
                models.update({model._meta.label_lower: model})
        return models



from __future__ import annotations

from typing import TYPE_CHECKING, Any, Self, Tuple

from ...constants import NOT_REQUIRED, REQUIRED
from ...metadata_updater import MetadataUpdater
from ..rule_group import RuleGroup, RuleGroupError, TargetModelConflict
from ..rule_group_metaclass import RuleGroupMetaclass

if TYPE_CHECKING:
    from edc_visit_schedule import FormsCollection
    from edc_visit_tracking.model_mixins import VisitModelMixin as Base

    from ...model_mixins.creates import CreatesMetadataModelMixin
    from ...models import CrfMetadata

    class RelatedVisitModel(CreatesMetadataModelMixin, Base):
        pass


class CrfRuleGroup(RuleGroup, metaclass=RuleGroupMetaclass):
    """A class used to declare and contain rules."""

    metadata_updater_cls = MetadataUpdater

    def __str__(self: Self) -> str:
        return f"{self.__class__.__name__}({self.name})"

    def __repr__(self: Self) -> str:
        return f"{self.__class__.__name__}({self.name})"

    @classmethod
    def crfs_for_visit(cls, visit: RelatedVisitModel = None) -> FormsCollection:
        """Returns a list of scheduled or unscheduled
        CRFs + PRNs depending on visit_code_sequence.
        """
        if visit.visit_code_sequence != 0:
            crfs = visit.visit.crfs_unscheduled + visit.visit.crfs_prn
        else:
            crfs = visit.visit.crfs + visit.visit.crfs_prn
        return crfs

    @classmethod
    def evaluate_rules(
        cls: Any,
        related_visit: RelatedVisitModel = None,
        allow_create: bool | None = None,
    ) -> Tuple[dict[str, dict[str, dict]], dict[str, CrfMetadata]]:
        rule_results = {}
        metadata_objects = {}
        for rule in cls._meta.options.get("rules"):
            # skip if source model is not in visit.crfs (including PRNs)
            if (
                rule.source_model
                and rule.source_model != related_visit._meta.label_lower
                and rule.source_model
                not in [c.model for c in cls.crfs_for_visit(related_visit)]
            ):
                continue
            for target_model in rule.target_models:
                if target_model == related_visit._meta.label_lower:
                    raise TargetModelConflict(
                        f"Target model and visit model are the same! "
                        f"Got {target_model}=={related_visit._meta.label_lower}"
                    )
                elif (
                    target_model.split(".")[1] == related_visit._meta.label_lower.split(".")[1]
                ):
                    raise TargetModelConflict(
                        f"Target model and visit model might be the same. "
                        f"Got {target_model}~={related_visit._meta.label_lower}"
                    )
            if result := rule.run(related_visit=related_visit):
                rule_results.update({str(rule): result})
                for target_model, entry_status in rule_results[str(rule)].items():
                    if not entry_status:
                        raise RuleGroupError("Cannot be None. Got `entry_status`.")
                    # only do something if target model is in visit.crfs (including PRNs)
                    if target_model in [c.model for c in cls.crfs_for_visit(related_visit)]:
                        metadata_updater = cls.metadata_updater_cls(
                            related_visit=related_visit,
                            target_model=target_model,
                            allow_create=allow_create,
                        )
                        metadata_obj = metadata_updater.get_and_update(
                            entry_status=entry_status
                        )
                        metadata_objects.update({target_model: metadata_obj})
        return rule_results, metadata_objects

    @classmethod
    def default_entry_status(
        cls, related_visit: RelatedVisitModel, target_model: Any
    ) -> str | None:
        """Returns the default `entry_status` or None"""
        for c in cls.crfs_for_visit(related_visit):
            if c.model == target_model:
                return REQUIRED if c.required else NOT_REQUIRED
        return None



from django.test import TestCase
from edc_appointment.models import Appointment
from edc_facility import import_holidays
from edc_lab.models import Panel
from edc_reference import site_reference_configs
from edc_utils import get_utcnow
from edc_visit_schedule import site_visit_schedules

from ...models import CrfMetadata, RequisitionMetadata
from ..models import SubjectConsent
from ..reference_configs import register_to_site_reference_configs
from ..visit_schedule import visit_schedule


class TestMetadataMixin(TestCase):
    @classmethod
    def setUpTestData(cls):
        import_holidays()

    def setUp(self):
        self.panel_one = Panel.objects.create(name="one")
        self.panel_two = Panel.objects.create(name="two")

        for name in ["three", "four", "five", "six"]:
            Panel.objects.create(name=name)

        site_visit_schedules._registry = {}
        site_visit_schedules.loaded = False
        site_visit_schedules.register(visit_schedule)

        register_to_site_reference_configs()
        site_reference_configs.register_from_visit_schedule(
            visit_models={"edc_appointment.appointment": "edc_visit_tracking.subjectvisit"}
        )

        self.subject_identifier = "1111111"
        self.assertEqual(CrfMetadata.objects.all().count(), 0)
        self.assertEqual(RequisitionMetadata.objects.all().count(), 0)
        subject_consent = SubjectConsent.objects.create(
            subject_identifier=self.subject_identifier, consent_datetime=get_utcnow()
        )
        _, self.schedule = site_visit_schedules.get_by_onschedule_model(
            "edc_metadata.onschedule"
        )
        self.schedule.put_on_schedule(
            subject_identifier=self.subject_identifier,
            onschedule_datetime=subject_consent.consent_datetime,
        )
        self.appointment = Appointment.objects.get(
            subject_identifier=self.subject_identifier,
            visit_code=self.schedule.visits.first.code,
        )

        self.appointment_2000 = Appointment.objects.get(
            subject_identifier=self.subject_identifier,
            visit_code="2000",
        )



from django.test import TestCase
from edc_constants.constants import MALE
from edc_facility.import_holidays import import_holidays

from edc_metadata import NOT_REQUIRED, REQUIRED

from ...metadata_rules import (
    CrfRule,
    CrfRuleGroup,
    P,
    RegisterRuleGroupError,
    SiteMetadataNoRulesError,
    SiteMetadataRulesAlreadyRegistered,
    register,
    site_metadata_rules,
)
from ..reference_configs import register_to_site_reference_configs


class RuleGroupWithoutRules(CrfRuleGroup):
    class Meta:
        app_label = "edc_metadata"
        source_model = "edc_visit_tracking.subjectvisit"


class RuleGroupWithRules(CrfRuleGroup):
    rule1 = CrfRule(
        predicate=P("gender", "eq", MALE),
        consequence=REQUIRED,
        alternative=NOT_REQUIRED,
        target_models=["crfone", "crftwo"],
    )

    class Meta:
        app_label = "edc_metadata"
        source_model = "edc_visit_tracking.subjectvisit"


class RuleGroupWithRules2(CrfRuleGroup):
    rule1 = CrfRule(
        predicate=P("gender", "eq", MALE),
        consequence=REQUIRED,
        alternative=NOT_REQUIRED,
        target_models=["crfone", "crftwo"],
    )

    class Meta:
        app_label = "edc_metadata"
        source_model = "edc_visit_tracking.subjectvisit"


class TestSiteMetadataRules(TestCase):
    @classmethod
    def setUpClass(cls):
        import_holidays()
        return super().setUpClass()

    def setUp(self):
        register_to_site_reference_configs()
        site_metadata_rules.registry = {}

    def test_register_rule_group_no_rules_raises_on_register(self):
        self.assertRaises(
            SiteMetadataNoRulesError,
            site_metadata_rules.register,
            RuleGroupWithoutRules,
        )

    def test_register_rule_group_with_rule(self):
        try:
            site_metadata_rules.register(RuleGroupWithRules)
        except SiteMetadataNoRulesError:
            self.fail("SiteMetadataNoRulesError unexpectedly raised.")

    def test_register_rule_group_get_rule_groups_for_app_label(self):
        site_metadata_rules.register(RuleGroupWithRules)
        rule_groups = site_metadata_rules.rule_groups.get("edc_metadata")
        self.assertEqual(rule_groups, [RuleGroupWithRules])

    def test_register_rule_group_register_more_than_one_rule_group(self):
        site_metadata_rules.register(RuleGroupWithRules)
        site_metadata_rules.register(RuleGroupWithRules2)
        rule_groups = site_metadata_rules.rule_groups.get("edc_metadata")
        self.assertEqual(rule_groups, [RuleGroupWithRules, RuleGroupWithRules2])

    def test_register_twice_raises(self):
        site_metadata_rules.register(rule_group_cls=RuleGroupWithRules)
        self.assertRaises(
            SiteMetadataRulesAlreadyRegistered,
            site_metadata_rules.register,
            RuleGroupWithRules,
        )

    def test_rule_group_repr(self):
        repr(RuleGroupWithRules())
        str(RuleGroupWithRules())

    def test_register_decorator(self):
        @register()
        class RuleGroupWithRules(CrfRuleGroup):
            rule1 = CrfRule(
                predicate=P("gender", "eq", MALE),
                consequence=REQUIRED,
                alternative=NOT_REQUIRED,
                target_models=["crfone", "crftwo"],
            )

            class Meta:
                app_label = "edc_metadata"
                source_model = "edc_visit_tracking.subjectvisit"

        self.assertIn("edc_metadata", site_metadata_rules.registry)

    def test_register_decorator_raises(self):
        try:

            @register()
            class RuleGroupWithRules:
                class Meta:
                    app_label = "edc_metadata"
                    source_model = "edc_visit_tracking.subjectvisit"

        except RegisterRuleGroupError:
            pass
        else:
            self.fail("RegisterRuleGroupError unexpectedly not raised.")



from .constants import KEYED, REQUIRED
from .models import CrfMetadata


class MetaDataInspector:

    """Inspects for the given timepoint and form."""

    metadata_model_cls = CrfMetadata

    def __init__(
        self,
        model_cls=None,
        visit_schedule_name=None,
        schedule_name=None,
        visit_code=None,
        timepoint=None,
    ):
        self.model_cls = model_cls
        self.visit_schedule_name = visit_schedule_name
        self.schedule_name = schedule_name
        self.visit_code = visit_code
        self.timepoint = timepoint

    def __repr__(self):
        return (
            f"{self.__class__.__name__}(<{self.model_cls._meta.label_lower} "
            f"{self.visit_schedule_name}."
            f"{self.schedule_name}.{self.visit_code}"
            f"@{self.timepoint}>)"
        )

    @property
    def required(self):
        """Returns subject_identifiers as a values queryset."""
        opts = dict(
            visit_schedule_name=self.visit_schedule_name,
            schedule_name=self.schedule_name,
            visit_code=self.visit_code,
            timepoint=self.timepoint,
            model=self.model_cls._meta.label_lower,
            entry_status=REQUIRED,
        )
        return (
            self.metadata_model_cls.objects.filter(**opts)
            .values("subject_identifier")
            .order_by("subject_identifier")
            .distinct()
        )

    @property
    def keyed(self):
        """Returns subject_identifiers as a values queryset."""
        opts = dict(
            visit_schedule_name=self.visit_schedule_name,
            schedule_name=self.schedule_name,
            visit_code=self.visit_code,
            timepoint=self.timepoint,
            model=self.model_cls._meta.label_lower,
            entry_status=KEYED,
        )
        return (
            self.metadata_model_cls.objects.filter(**opts)
            .values("subject_identifier")
            .order_by("subject_identifier")
            .distinct()
        )



from dateutil.relativedelta import relativedelta
from django.contrib.sites.models import Site
from edc_utils import get_utcnow
from edc_visit_tracking.constants import SCHEDULED
from faker import Faker
from model_bakery.recipe import Recipe, seq

from .models import SubjectConsent, SubjectVisit

fake = Faker()

subjectvisit = Recipe(SubjectVisit, reason=SCHEDULED)

subjectconsent = Recipe(
    SubjectConsent,
    confirm_identity=seq("12315678"),
    consent_datetime=get_utcnow(),
    dob=get_utcnow() - relativedelta(years=25),
    first_name=fake.first_name,
    gender="M",
    identity=seq("12315678"),
    initials="XX",
    is_dob_estimated="-",
    last_name=fake.last_name,
    screening_identifier=None,
    site=Site.objects.get_current(),
    subject_identifier=None,
    user_created="erikvw",
    user_modified="erikvw",
)



import sys
from warnings import warn

from django.core.management import CommandError
from django.core.management.base import BaseCommand
from django.core.management.color import color_style
from edc_reference.models import Reference

from edc_metadata.metadata_refresher import MetadataRefresher
from edc_metadata.models import CrfMetadata, RequisitionMetadata

style = color_style()


class Command(BaseCommand):
    help = "Update references, metadata and re-run metadatarules"

    def add_arguments(self, parser):
        parser.add_argument(
            "--update_metadata_only",
            dest="update_metadata_only",
            default=False,
            action="store_true",
            help="Update metadata only, skip references (NOT RECOMMENDED)",
        )
        parser.add_argument(
            "--update_references_only",
            dest="update_references_only",
            default=False,
            action="store_true",
            help="Update references only, skip medata. (NOT RECOMMENDED)",
        )

        parser.add_argument(
            "--delete_metadata",
            dest="delete_metadata",
            default=False,
            action="store_true",
            help="delete existing metadata first",
        )

        parser.add_argument(
            "--delete_references",
            dest="delete_references",
            default=False,
            action="store_true",
            help="delete existing references first",
        )

    def handle(self, *args, **options) -> None:
        update_metadata_only = options.get("update_metadata_only")
        update_references_only = options.get("update_references_only")
        delete_metadata = options.get("delete_metadata")
        delete_references = options.get("delete_references")
        if update_metadata_only and update_references_only:
            raise CommandError(
                "Invalid options. Use `update_metadata_only` or `update_references_only` "
                "but not both."
            )
        elif update_metadata_only or update_references_only:
            warn(
                "Option not recommended. Updating metadata and references "
                "separately may lead to unexpected results."
            )
        metadata_refresher = MetadataRefresher(verbose=True)
        if delete_metadata:
            sys.stdout.write("Deleting all CrfMetadata...     \r")
            CrfMetadata.objects.all().delete()
            sys.stdout.write("Deleting all CrfMetadata...done.                    \n")
            sys.stdout.write("Deleting all RequisitionMetadata...     \r")
            RequisitionMetadata.objects.all().delete()
            sys.stdout.write("Deleting all RequisitionMetadata...done.            \n")
        if delete_references:
            sys.stdout.write("Deleting all References...     \r")
            Reference.objects.all().delete()
            sys.stdout.write("Deleting all References...done.                    \n")
        if update_metadata_only:
            sys.stdout.write("Updating metadata for all post consent models ...     \n")
            sys.stdout.write("  Note: References will not be updated;\n")
            sys.stdout.write("        Metadata rules will not be run.\n\n")
            metadata_refresher.create_or_update_metadata_for_all()
        elif update_references_only:
            sys.stdout.write("Updating references for all source models ...     \n")
            sys.stdout.write("  Note: Metadata will not be updated.\n\n")
            metadata_refresher.create_or_update_references_for_all()
        else:
            metadata_refresher.run()



from dateutil.relativedelta import relativedelta
from django.core.exceptions import ObjectDoesNotExist
from django.test import TestCase
from edc_visit_schedule import (
    Crf,
    FormsCollection,
    Schedule,
    Visit,
    VisitSchedule,
    site_visit_schedules,
)
from edc_visit_tracking.constants import SCHEDULED

from edc_metadata.constants import KEYED, REQUIRED
from edc_metadata.metadata_refresher import MetadataRefresher
from edc_metadata.models import CrfMetadata

from ..models import CrfFive, CrfOne, SubjectVisit
from .metadata_test_mixin import TestMetadataMixin


class TestMetadataRefresher(TestMetadataMixin, TestCase):
    def check(self, expected, subject_visit=None):
        for model, entry_status in expected.items():
            with self.subTest(model=model, entry_status=entry_status):
                try:
                    CrfMetadata.objects.filter(
                        entry_status=KEYED,
                        model="edc_metadata.crfone",
                        visit_code=subject_visit.visit_code,
                    )
                except ObjectDoesNotExist:
                    self.fail(
                        f"CrfMetadata unexpectedly does not exist. Got {model}, {entry_status}"
                    )

    def test_updates_crf_metadata_as_keyed(self):
        subject_visit = SubjectVisit.objects.create(
            appointment=self.appointment,
            visit_code=self.appointment.visit_code,
            visit_code_sequence=self.appointment.visit_code_sequence,
            visit_schedule_name=self.appointment.visit_schedule_name,
            schedule_name=self.appointment.schedule_name,
            reason=SCHEDULED,
        )
        CrfOne.objects.create(subject_visit=subject_visit)
        CrfMetadata.objects.all().delete()
        expected = {
            "edc_metadata.crfone": KEYED,
            "edc_metadata.crftwo": REQUIRED,
            "edc_metadata.crfthree": REQUIRED,
        }
        metadata_refresher = MetadataRefresher()
        metadata_refresher.run()
        self.check(expected, subject_visit=subject_visit)

    def test_updates_after_manual_change(self):
        subject_visit = SubjectVisit.objects.create(
            appointment=self.appointment,
            visit_code=self.appointment.visit_code,
            visit_code_sequence=self.appointment.visit_code_sequence,
            visit_schedule_name=self.appointment.visit_schedule_name,
            schedule_name=self.appointment.schedule_name,
            reason=SCHEDULED,
        )
        CrfOne.objects.create(subject_visit=subject_visit)
        CrfMetadata.objects.all().delete()
        expected = {
            "edc_metadata.crfone": KEYED,
            "edc_metadata.crftwo": REQUIRED,
            "edc_metadata.crfthree": REQUIRED,
        }
        metadata_refresher = MetadataRefresher()
        metadata_refresher.run()
        crf_metadata = CrfMetadata.objects.get(model="edc_metadata.crfone")
        crf_metadata.entry_status = REQUIRED
        crf_metadata.save()
        metadata_refresher = MetadataRefresher()
        metadata_refresher.run()
        self.check(expected, subject_visit=subject_visit)

    def test_updates_after_manual_change2(self):
        subject_visit = SubjectVisit.objects.create(
            appointment=self.appointment,
            visit_code=self.appointment.visit_code,
            visit_code_sequence=self.appointment.visit_code_sequence,
            visit_schedule_name=self.appointment.visit_schedule_name,
            schedule_name=self.appointment.schedule_name,
            reason=SCHEDULED,
        )
        crf_one = CrfOne.objects.create(subject_visit=subject_visit)
        CrfMetadata.objects.all().delete()
        metadata_refresher = MetadataRefresher()
        metadata_refresher.run()
        expected = {
            "edc_metadata.crfone": REQUIRED,
            "edc_metadata.crftwo": REQUIRED,
            "edc_metadata.crfthree": REQUIRED,
        }
        crf_metadata = CrfMetadata.objects.get(model="edc_metadata.crfone")
        crf_one.delete()
        crf_metadata.entry_status = KEYED
        crf_metadata.save()
        metadata_refresher = MetadataRefresher()
        metadata_refresher.run()
        self.check(expected, subject_visit=subject_visit)

    @staticmethod
    def register_new_visit_schedule(crfs):
        site_visit_schedules._registry = {}
        site_visit_schedules.loaded = False

        visit = Visit(
            code="1000",
            title="Week 1",
            timepoint=0,
            rbase=relativedelta(days=0),
            rlower=relativedelta(days=0),
            rupper=relativedelta(days=6),
            crfs=crfs,
            allow_unscheduled=True,
            facility_name="5-day-clinic",
        )
        schedule = Schedule(
            name="schedule",
            onschedule_model="edc_metadata.onschedule",
            offschedule_model="edc_metadata.offschedule",
            consent_model="edc_metadata.subjectconsent",
            appointment_model="edc_appointment.appointment",
        )
        schedule.add_visit(visit)
        new_visit_schedule = VisitSchedule(
            name="visit_schedule",
            offstudy_model="edc_offstudy.subjectoffstudy",
            death_report_model="edc_metadata.deathreport",
        )
        new_visit_schedule.add_schedule(schedule)
        site_visit_schedules.register(new_visit_schedule)

    def test_after_schedule_change(self):
        subject_visit = SubjectVisit.objects.create(
            appointment=self.appointment,
            visit_code=self.appointment.visit_code,
            visit_code_sequence=self.appointment.visit_code_sequence,
            visit_schedule_name=self.appointment.visit_schedule_name,
            schedule_name=self.appointment.schedule_name,
            reason=SCHEDULED,
        )
        CrfFive.objects.create(subject_visit=subject_visit)
        self.assertEqual(CrfMetadata.objects.all().count(), 5)
        crfs = FormsCollection(
            Crf(show_order=1, model="edc_metadata.crfone", required=True),
            Crf(show_order=2, model="edc_metadata.crftwo", required=True),
            Crf(show_order=3, model="edc_metadata.crfthree", required=True),
            Crf(show_order=4, model="edc_metadata.crffour", required=True),
        )
        self.register_new_visit_schedule(crfs)
        self.assertEqual(CrfMetadata.objects.all().count(), 5)
        metadata_refresher = MetadataRefresher()
        metadata_refresher.run()
        # CrfFive is no longer in visit schedule, so one less
        # CRFMetadata instance
        self.assertEqual(CrfMetadata.objects.all().count(), 4)
        expected = {
            "edc_metadata.crfone": REQUIRED,
            "edc_metadata.crftwo": REQUIRED,
            "edc_metadata.crfthree": REQUIRED,
            "edc_metadata.crfFour": REQUIRED,
        }
        self.check(expected, subject_visit=subject_visit)

    def test_after_schedule_change2(self):
        subject_visit = SubjectVisit.objects.create(
            appointment=self.appointment,
            subject_identifier=self.subject_identifier,
            report_datetime=self.appointment.appt_datetime,
            visit_code=self.appointment.visit_code,
            visit_code_sequence=self.appointment.visit_code_sequence,
            visit_schedule_name=self.appointment.visit_schedule_name,
            schedule_name=self.appointment.schedule_name,
            reason=SCHEDULED,
        )
        self.assertEqual(CrfMetadata.objects.all().count(), 5)
        crfs = FormsCollection(
            Crf(show_order=1, model="edc_metadata.crfone", required=True),
            Crf(show_order=2, model="edc_metadata.crftwo", required=True),
            Crf(show_order=3, model="edc_metadata.crfthree", required=True),
            Crf(show_order=4, model="edc_metadata.crffour", required=True),
        )
        self.register_new_visit_schedule(crfs)
        self.assertEqual(CrfMetadata.objects.all().count(), 5)
        metadata_refresher = MetadataRefresher()
        metadata_refresher.run()
        self.assertEqual(CrfMetadata.objects.all().count(), 4)
        expected = {
            "edc_metadata.crfone": REQUIRED,
            "edc_metadata.crftwo": REQUIRED,
            "edc_metadata.crfthree": REQUIRED,
            "edc_metadata.crfFour": REQUIRED,
        }
        self.check(expected, subject_visit=subject_visit)



import sys

from django.apps.config import AppConfig as DjangoAppConfig
from django.core.management.color import color_style

from .metadata_rules import site_metadata_rules

style = color_style()


class AppConfig(DjangoAppConfig):
    name = "edc_metadata"
    verbose_name = "Data collection status"
    metadata_rules_enabled = True
    has_exportable_data = True
    include_in_administration_section = True

    def ready(self):
        sys.stdout.write(f"Loading {self.name} ...\n")
        site_metadata_rules.autodiscover()
        if not site_metadata_rules.registry:
            sys.stdout.write(
                style.ERROR(" Warning. No metadata rules were loaded by autodiscover.\n")
            )
        if not self.metadata_rules_enabled:
            sys.stdout.write(style.NOTICE(" * metadata rules disabled!\n"))
        sys.stdout.write(f" Done loading {self.name}.\n")



from __future__ import annotations

from typing import TYPE_CHECKING, Type

from django.apps import apps as django_apps

from .constants import CRF, KEYED
from .metadata_handler import MetadataHandler

if TYPE_CHECKING:
    from edc_visit_tracking.model_mixins import VisitModelMixin as Base

    from .model_mixins.creates import CreatesMetadataModelMixin
    from .models import CrfMetadata, RequisitionMetadata

    class RelatedVisitModel(CreatesMetadataModelMixin, Base):
        pass


class MetadataUpdaterError(Exception):
    pass


class MetadataUpdater:
    """A class to update a subject's metadata given
    the related_visit, target model name and desired entry status.
    """

    metadata_handler_cls: Type[MetadataHandler] = MetadataHandler
    metadata_category: str = CRF
    metadata_model: str = "edc_metadata.crfmetadata"

    def __init__(
        self,
        related_visit: RelatedVisitModel = None,
        target_model: str = None,
        allow_create: bool | None = None,
    ):
        self._metadata_obj: CrfMetadata | RequisitionMetadata | None = None
        self.allow_create = True if allow_create is None else allow_create
        self.related_visit = related_visit
        self.target_model = target_model

    def __repr__(self):
        return (
            f"{self.__class__.__name__}"
            f"(related_visit={self.related_visit}, "
            f"target_model={self.target_model})"
        )

    def get_and_update(self, entry_status: str = None) -> CrfMetadata | RequisitionMetadata:
        metadata_obj = self.metadata_handler.metadata_obj
        if entry_status != KEYED and self.is_keyed(metadata_obj.model):
            entry_status = KEYED
        if metadata_obj.entry_status != entry_status:
            metadata_obj.entry_status = entry_status
            metadata_obj.save(update_fields=["entry_status"])
            metadata_obj.refresh_from_db()
            if metadata_obj.entry_status != entry_status:
                raise MetadataUpdaterError(
                    "Expected entry status does not match `entry_status` on "
                    "metadata model instance. "
                    f"Got {entry_status} != {metadata_obj.entry_status}."
                )
        return metadata_obj

    def is_keyed(self, model: str) -> bool:
        """Returns True if source model exists."""
        return (
            django_apps.get_model(model)
            .objects.filter(subject_visit=self.related_visit)
            .exists()
        )

    @property
    def metadata_handler(self) -> MetadataHandler:
        return self.metadata_handler_cls(
            metadata_model=self.metadata_model,
            model=self.target_model,
            related_visit=self.related_visit,
            allow_create=self.allow_create,
        )



# Generated by Django 2.2.2 on 2019-07-08 22:09

import sys

from django.db import migrations
from edc_visit_schedule.site_visit_schedules import site_visit_schedules


def update_metadata_timepoint(apps, schema_editor):
    CrfMetadata = apps.get_model("edc_metadata", "CrfMetadata")
    RequisitionMetadata = apps.get_model("edc_metadata", "RequisitionMetadata")
    for visit_schedule in site_visit_schedules.visit_schedules.values():
        for schedule in visit_schedule.schedules.values():
            for visit in schedule.visits.values():
                visit_str = (
                    f"{visit_schedule.name}.{schedule.name}.{visit.code}@" f"{visit.timepoint}"
                )
                sys.stdout.write(f"  - updating CRF metadata timepoint for {visit_str}\n")
                CrfMetadata.objects.filter(
                    visit_schedule_name=visit_schedule.name,
                    schedule_name=schedule.name,
                    visit_code=visit.code,
                    timepoint__isnull=True,
                ).update(timepoint=visit.timepoint)
                sys.stdout.write(
                    f"  - updating Requisition metadata timepoint for {visit_str}\n"
                )
                RequisitionMetadata.objects.filter(
                    visit_schedule_name=visit_schedule.name,
                    schedule_name=schedule.name,
                    visit_code=visit.code,
                    timepoint__isnull=True,
                ).update(timepoint=visit.timepoint)


class Migration(migrations.Migration):
    dependencies = [("edc_metadata", "0014_auto_20190707_0002")]

    operations = [migrations.RunPython(update_metadata_timepoint)]



from __future__ import annotations

from django.apps import apps as django_apps
from django.contrib import messages
from django.core.exceptions import ObjectDoesNotExist
from django.utils.translation import gettext as _
from edc_appointment.constants import IN_PROGRESS_APPT
from edc_subject_model_wrappers import CrfModelWrapper, RequisitionModelWrapper

from ..constants import CRF, KEYED, NOT_REQUIRED, REQUIRED, REQUISITION
from ..metadata_wrappers import CrfMetadataWrappers, RequisitionMetadataWrappers
from ..utils import refresh_references_and_metadata_for_timepoint


class MetadataViewError(Exception):
    pass


class MetadataViewMixin:
    crf_model_wrapper_cls: CrfModelWrapper = CrfModelWrapper
    requisition_model_wrapper_cls: RequisitionModelWrapper = RequisitionModelWrapper
    crf_metadata_wrappers_cls: CrfMetadataWrappers = CrfMetadataWrappers
    requisition_metadata_wrappers_cls: RequisitionMetadataWrappers = (
        RequisitionMetadataWrappers
    )
    panel_model: str = "edc_lab.panel"

    metadata_show_status: list[str] = [REQUIRED, KEYED]

    def get(self, request, *args, **kwargs):
        try:
            referrer = request.headers.get("Referer")
        except TypeError:
            pass
        else:
            if (
                referrer
                and kwargs.get("appointment")
                and "subject_review_listboard" in referrer
            ):
                self.appointment_id = kwargs.get("appointment")
                self.refresh_references_and_metadata_for_timepoint()
        return super().get(request, *args, **kwargs)

    def get_context_data(self, **kwargs):
        context = super().get_context_data(**kwargs)
        context.update(metadata_show_status=self.metadata_show_status)
        if self.appointment:
            if self.appointment.appt_status != IN_PROGRESS_APPT:
                message = _(
                    'You have selected an appointment that is no longer "in progress". '
                    "Refer to the schedule for the appointment that is "
                    'currently "in progress".'
                )
                self.message_user(message, level=messages.WARNING)

            crf_model_wrappers = self.get_crf_model_wrappers()
            try:
                report_datetime = self.appointment.related_visit.report_datetime
            except AttributeError:
                pass
            else:
                context.update(
                    report_datetime=report_datetime,
                    crfs=crf_model_wrappers,
                    requisitions=self.get_requisition_model_wrapper(),
                    NOT_REQUIRED=NOT_REQUIRED,
                    REQUIRED=REQUIRED,
                    KEYED=KEYED,
                )
        return context

    def refresh_references_and_metadata_for_timepoint(self):
        """Save related visit model instance to run metadata update."""
        refresh_references_and_metadata_for_timepoint(self.appointment, skip_references=True)

    def get_crf_model_wrappers(self):
        """Returns a list of model wrappers.

        Gets each CrfMetadata instance, validates the entry status and wraps
        in a model wrapper.
        """
        model_wrappers = []
        crf_metadata_wrappers = self.crf_metadata_wrappers_cls(appointment=self.appointment)
        for metadata_wrapper in crf_metadata_wrappers.objects:
            if not metadata_wrapper.model_obj:
                metadata_wrapper.model_obj = metadata_wrapper.model_cls(
                    **{
                        metadata_wrapper.model_cls.related_visit_model_attr(): (
                            metadata_wrapper.visit
                        )
                    }
                )
            metadata_wrapper.metadata_obj.object = self.crf_model_wrapper_cls(
                model_obj=metadata_wrapper.model_obj,
                model=metadata_wrapper.metadata_obj.model,
                key=CRF,
                request=self.request,
            )
            model_wrappers.append(metadata_wrapper.metadata_obj)
        return [
            model_wrapper
            for model_wrapper in model_wrappers
            if model_wrapper.entry_status in self.metadata_show_status
        ]

    def get_requisition_model_wrapper(self):
        """Returns a list of model wrappers."""
        model_wrappers = []
        requisition_metadata_wrappers = self.requisition_metadata_wrappers_cls(
            appointment=self.appointment
        )
        for metadata_wrapper in requisition_metadata_wrappers.objects:
            if not metadata_wrapper.model_obj:
                panel = self.get_panel(metadata_wrapper)
                metadata_wrapper.model_obj = metadata_wrapper.model_cls(
                    **{
                        metadata_wrapper.model_cls.related_visit_model_attr(): (
                            metadata_wrapper.visit
                        ),
                        "panel": panel,
                    }
                )
            metadata_wrapper.metadata_obj.object = self.requisition_model_wrapper_cls(
                model_obj=metadata_wrapper.model_obj,
                model=metadata_wrapper.metadata_obj.model,
                key=REQUISITION,
                request=self.request,
            )
            model_wrappers.append(metadata_wrapper.metadata_obj)
        return [
            model_wrapper
            for model_wrapper in model_wrappers
            if model_wrapper.entry_status in self.metadata_show_status
        ]

    def get_panel(self, metadata_wrapper=None):
        try:
            panel = self.panel_model_cls.objects.get(name=metadata_wrapper.panel_name)
        except ObjectDoesNotExist as e:
            raise MetadataViewError(
                f"{e} Got panel name '{metadata_wrapper.panel_name}'. "
                f"See {metadata_wrapper}."
            )
        return panel

    @property
    def panel_model_cls(self):
        return django_apps.get_model(self.panel_model)



from __future__ import annotations

from typing import TYPE_CHECKING, Any

from django.apps import apps as django_apps
from django.core.exceptions import ObjectDoesNotExist
from django.db import models
from edc_identifier.model_mixins import NonUniqueSubjectIdentifierFieldMixin
from edc_sites.model_mixins import SiteModelMixin
from edc_visit_schedule.model_mixins import (
    VisitScheduleFieldsModelMixin,
    VisitScheduleMethodsModelMixin,
)

from ..choices import ENTRY_STATUS, NOT_REQUIRED, REQUIRED
from ..constants import KEYED

if TYPE_CHECKING:
    from django.contrib.sites.models import Site


class CrfMetadataModelMixin(
    NonUniqueSubjectIdentifierFieldMixin,
    SiteModelMixin,
    VisitScheduleMethodsModelMixin,
    VisitScheduleFieldsModelMixin,
    models.Model,
):

    """Mixin for CrfMetadata and RequisitionMetadata models."""

    visit_code = models.CharField(max_length=25)

    visit_code_sequence = models.IntegerField(default=0)

    timepoint = models.DecimalField(null=True, decimal_places=1, max_digits=6)

    model = models.CharField(max_length=50)

    current_entry_title = models.CharField(max_length=250, null=True)

    show_order = models.IntegerField()  # must always be provided!

    entry_status = models.CharField(
        max_length=25, choices=ENTRY_STATUS, default=REQUIRED, db_index=True
    )

    due_datetime = models.DateTimeField(null=True, blank=True)

    report_datetime = models.DateTimeField(null=True, blank=True)

    entry_comment = models.TextField(max_length=250, null=True, blank=True)

    close_datetime = models.DateTimeField(null=True, blank=True)

    fill_datetime = models.DateTimeField(null=True, blank=True)

    def natural_key(self):
        return (
            self.subject_identifier,
            self.visit_schedule_name,
            self.schedule_name,
            self.visit_code,
            self.visit_code_sequence,
            self.model,
        )

    def is_required(self):
        return self.entry_status != NOT_REQUIRED

    def is_not_required(self):
        return not self.is_required()

    def model_instance_query_opts(self) -> dict:
        models_cls = django_apps.get_model(self.model)
        attr = models_cls.related_visit_model_attr()
        return {
            f"{attr}__schedule_name": self.schedule_name,
            f"{attr}__site": self.site,
            f"{attr}__subject_identifier": self.subject_identifier,
            f"{attr}__visit_code": self.visit_code,
            f"{attr}__visit_code_sequence": self.visit_code_sequence,
            f"{attr}__visit_schedule_name": self.visit_schedule_name,
        }

    @property
    def model_cls(self):
        return django_apps.get_model(self.model)

    @property
    def model_instance(self: Any) -> Any:
        """Returns the CRF/Requisition model instance or None"""
        instance = None
        try:
            instance = self.model_cls.objects.get(**self.model_instance_query_opts())
        except ObjectDoesNotExist:
            pass
        return instance

    def refresh_entry_status(self) -> str:
        """Resets entry_status to the original visit schedule value"""
        if not self.model_instance:
            if self.visit_code_sequence > 0:
                self.entry_status = REQUIRED
            else:
                visit = self.visits.get(self.visit_code)
                for required in [
                    crf.required for crf in visit.crfs if crf.model == self.model
                ]:
                    self.entry_status = REQUIRED if required else NOT_REQUIRED
        else:
            self.entry_status = KEYED
        self.save_base(update_fields=["entry_status"])
        return self.entry_status

    def get_entry_status(self) -> str:
        return self.refresh_entry_status()

    def get_site_on_create(self) -> Site:
        """Expect site instance to be set from the reference model
        instance.
        """
        return self.site

    class Meta:
        abstract = True
        ordering = ("show_order",)



from __future__ import annotations

from typing import Any, Optional, Union

from edc_reference.reference import ReferenceGetterError, ReferenceObjectDoesNotExist


class PredicateError(Exception):
    pass


class NoValueError(Exception):
    pass


class BasePredicate:
    def get_value(
        self,
        attr: str = None,
        source_model: str | None = None,
        reference_getter_cls: Optional[Any] = None,
        **kwargs,
    ) -> Any:
        """Returns a value by checking for the attr on each arg.

        Each arg in args may be a model instance, queryset, or None.

        A NoValueError is raised if attr is not found on any "instance".
        in kwargs.
        """
        found_on_instance: Any = None
        for instance in kwargs.values():
            try:
                getattr(instance, attr)
            except AttributeError:
                pass
            else:
                found_on_instance = instance
                break
        if found_on_instance:
            value = getattr(found_on_instance, attr)
        else:
            opts = dict(
                field_name=attr, name=source_model, **self.opts_from_visit(kwargs.get("visit"))
            )
            try:
                reference = reference_getter_cls(**opts)
            except (ReferenceGetterError, ReferenceObjectDoesNotExist) as e:
                raise NoValueError(f"No value found for {attr}. Given {kwargs}. Got {e}.")
            else:
                if reference.has_value:
                    value = getattr(reference, attr)
                else:
                    raise NoValueError(f"No value found for {attr}. Given {kwargs}")
        return value

    @staticmethod
    def opts_from_visit(visit: Any) -> dict:
        """Returns a dict of values from the visit model instance"""
        try:
            opts = dict(
                subject_identifier=visit.subject_identifier,
                report_datetime=visit.report_datetime,
                visit_schedule_name=visit.visit_schedule_name,
                schedule_name=visit.schedule_name,
                visit_code=visit.visit_code,
                visit_code_sequence=visit.visit_code_sequence,
                timepoint=visit.timepoint,
                site=visit.site,
            )
        except AttributeError as e:
            raise PredicateError(f"Invalid visit model or None. Got {e}")
        return opts


class P(BasePredicate):

    """
    Simple predicate class.

    For example:

        predicate = P('gender', 'eq', 'MALE')
        predicate = P('referral_datetime', 'is not', None)
        predicate = P('age', '<=', 64)
    """

    funcs = {
        "is": lambda x, y: True if x is y else False,
        "is not": lambda x, y: True if x is not y else False,
        "gt": lambda x, y: True if x > y else False,
        ">": lambda x, y: True if x > y else False,
        "gte": lambda x, y: True if x >= y else False,
        ">=": lambda x, y: True if x >= y else False,
        "lt": lambda x, y: True if x < y else False,
        "<": lambda x, y: True if x < y else False,
        "lte": lambda x, y: True if x <= y else False,
        "<=": lambda x, y: True if x <= y else False,
        "eq": lambda x, y: True if x == y else False,
        "equals": lambda x, y: True if x == y else False,
        "==": lambda x, y: True if x == y else False,
        "neq": lambda x, y: True if x != y else False,
        "!=": lambda x, y: True if x != y else False,
        "in": lambda x, y: True if x in y else False,
    }

    def __init__(self, attr: str, operator: str, expected_value: Union[list, str]) -> None:
        self.attr = attr
        self.expected_value = expected_value
        self.func = self.funcs.get(operator)
        if not self.func:
            raise PredicateError(f"Invalid operator. Got {operator}.")
        self.operator = operator

    def __repr__(self) -> str:
        return (
            f"{self.__class__.__name__}({self.attr}, {self.operator}, "
            f"{self.expected_value})"
        )

    def __call__(self, **kwargs) -> bool:
        value = self.get_value(attr=self.attr, **kwargs)
        return self.func(value, self.expected_value)


class PF(BasePredicate):
    """
    Predicate with a lambda function.

    predicate = PF('age', lambda x: True if x >= 18 and x <= 64 else False)

    if lamda is anything more complicated just pass a func directly to the predicate attr:

        def my_func(visit, registered_subject, source_obj, source_qs):
            if source_obj.married and registered_subject.gender == FEMALE:
                return True
            return False

        class MyRuleGroup(RuleGroup):

            my_rule = Rule(
                ...
                predicate = my_func
                ...

    """

    def __init__(self, *attrs, func: callable = None) -> None:
        self.attrs = attrs
        self.func = func

    def __call__(self, **kwargs) -> Any:
        values = []
        for attr in self.attrs:
            values.append(self.get_value(attr=attr, **kwargs))
        return self.func(*values)

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}({self.attrs}, {self.func})"



from __future__ import annotations

from typing import TYPE_CHECKING

from edc_visit_tracking.constants import MISSED_VISIT

from ..metadata_handler import MetadataHandler, MetadataHandlerError

if TYPE_CHECKING:
    from edc_lab.models import Panel

    from ..models import RequisitionMetadata


class RequisitionMetadataHandler(MetadataHandler):

    """A class to get or create a requisition metadata
    model instance.
    """

    def __init__(self, panel: Panel = None, **kwargs):
        super().__init__(**kwargs)
        self.panel = panel

    def _create(self, exception_msg: str | None = None) -> RequisitionMetadata:
        """Returns a created RequisitionMetadata model instance for this
        requisition.
        """
        metadata_obj = None
        try:
            requisition_object = [
                requisition
                for requisition in self.creator.visit.all_requisitions
                if requisition.panel.name == self.panel.name
            ][0]
        except IndexError as e:
            if self.related_visit.reason != MISSED_VISIT:
                raise MetadataHandlerError(
                    f"Model not found. Not in visit.all_crfs. Model {self.model}. Got {e}"
                )
        else:
            metadata_obj = self.creator.create_requisition(requisition_object)
        return metadata_obj

    @property
    def query_options(self) -> dict:
        """Returns a dict of options to query the metadata model."""
        query_options = super().query_options
        query_options.update({"panel_name": self.panel.name})
        return query_options



from .constants import CRF, DO_NOTHING, KEYED, NOT_REQUIRED, REQUIRED, REQUISITION
from .metadata_handler import MetadataObjectDoesNotExist
from .metadata_rules import (
    SiteMetadataNoRulesError,
    SiteMetadataRulesAlreadyRegistered,
    site_metadata_rules,
)
from .metadata_updater import MetadataUpdater
from .next_form_getter import NextFormGetter



# -*- coding: utf-8 -*-
# Generated by Django 1.10.3 on 2016-11-25 07:55
from __future__ import unicode_literals

import django_audit_fields.fields.uuid_auto_field
import django_extensions.db.fields
import django_revision.revision_field
import edc_model_fields.fields.hostname_modification_field
import edc_model_fields.fields.userfield
from django.db import migrations, models


class Migration(migrations.Migration):
    initial = True

    dependencies = []

    operations = [
        migrations.CreateModel(
            name="CrfMetadata",
            fields=[
                (
                    "created",
                    django_extensions.db.fields.CreationDateTimeField(
                        auto_now_add=True, verbose_name="created"
                    ),
                ),
                (
                    "modified",
                    django_extensions.db.fields.ModificationDateTimeField(
                        auto_now=True, verbose_name="modified"
                    ),
                ),
                (
                    "user_created",
                    edc_model_fields.fields.userfield.UserField(
                        blank=True,
                        editable=False,
                        max_length=50,
                        verbose_name="user created",
                    ),
                ),
                (
                    "user_modified",
                    edc_model_fields.fields.userfield.UserField(
                        blank=True,
                        editable=False,
                        max_length=50,
                        verbose_name="user modified",
                    ),
                ),
                (
                    "hostname_created",
                    models.CharField(
                        default="mac2-2.local",
                        editable=False,
                        help_text="System field. (modified on create only)",
                        max_length=50,
                    ),
                ),
                (
                    "hostname_modified",
                    edc_model_fields.fields.hostname_modification_field.HostnameModificationField(
                        blank=True,
                        editable=False,
                        help_text="System field. (modified on every save)",
                        max_length=50,
                    ),
                ),
                (
                    "revision",
                    django_revision.revision_field.RevisionField(
                        blank=True,
                        editable=False,
                        help_text="System field. Git repository tag:branch:commit.",
                        max_length=75,
                        null=True,
                        verbose_name="Revision",
                    ),
                ),
                (
                    "id",
                    django_audit_fields.fields.uuid_auto_field.UUIDAutoField(
                        blank=True,
                        editable=False,
                        help_text="System auto field. UUID primary key.",
                        primary_key=True,
                        serialize=False,
                    ),
                ),
                (
                    "subject_identifier",
                    models.CharField(
                        editable=False, max_length=50, verbose_name="Subject Identifier"
                    ),
                ),
                ("visit_schedule_name", models.CharField(max_length=25)),
                ("schedule_name", models.CharField(max_length=25)),
                ("visit_code", models.CharField(max_length=25)),
                ("model", models.CharField(max_length=50)),
                ("current_entry_title", models.CharField(max_length=250, null=True)),
                ("show_order", models.IntegerField()),
                (
                    "entry_status",
                    models.CharField(
                        choices=[
                            ("REQUIRED", "New"),
                            ("KEYED", "Keyed"),
                            ("MISSED", "Missed"),
                            ("NOT_REQUIRED", "Not required"),
                        ],
                        db_index=True,
                        default="REQUIRED",
                        max_length=25,
                    ),
                ),
                ("due_datetime", models.DateTimeField(blank=True, null=True)),
                ("report_datetime", models.DateTimeField(blank=True, null=True)),
                (
                    "entry_comment",
                    models.TextField(blank=True, max_length=250, null=True),
                ),
                ("close_datetime", models.DateTimeField(blank=True, null=True)),
                ("fill_datetime", models.DateTimeField(blank=True, null=True)),
            ],
            options={
                "verbose_name": "Crf Metadata",
                "verbose_name_plural": "Crf Metadata",
                "abstract": False,
            },
        ),
        migrations.CreateModel(
            name="RequisitionMetadata",
            fields=[
                (
                    "created",
                    django_extensions.db.fields.CreationDateTimeField(
                        auto_now_add=True, verbose_name="created"
                    ),
                ),
                (
                    "modified",
                    django_extensions.db.fields.ModificationDateTimeField(
                        auto_now=True, verbose_name="modified"
                    ),
                ),
                (
                    "user_created",
                    edc_model_fields.fields.userfield.UserField(
                        blank=True,
                        editable=False,
                        max_length=50,
                        verbose_name="user created",
                    ),
                ),
                (
                    "user_modified",
                    edc_model_fields.fields.userfield.UserField(
                        blank=True,
                        editable=False,
                        max_length=50,
                        verbose_name="user modified",
                    ),
                ),
                (
                    "hostname_created",
                    models.CharField(
                        default="mac2-2.local",
                        editable=False,
                        help_text="System field. (modified on create only)",
                        max_length=50,
                    ),
                ),
                (
                    "hostname_modified",
                    edc_model_fields.fields.hostname_modification_field.HostnameModificationField(
                        blank=True,
                        editable=False,
                        help_text="System field. (modified on every save)",
                        max_length=50,
                    ),
                ),
                (
                    "revision",
                    django_revision.revision_field.RevisionField(
                        blank=True,
                        editable=False,
                        help_text="System field. Git repository tag:branch:commit.",
                        max_length=75,
                        null=True,
                        verbose_name="Revision",
                    ),
                ),
                (
                    "id",
                    django_audit_fields.fields.uuid_auto_field.UUIDAutoField(
                        blank=True,
                        editable=False,
                        help_text="System auto field. UUID primary key.",
                        primary_key=True,
                        serialize=False,
                    ),
                ),
                (
                    "subject_identifier",
                    models.CharField(
                        editable=False, max_length=50, verbose_name="Subject Identifier"
                    ),
                ),
                ("visit_schedule_name", models.CharField(max_length=25)),
                ("schedule_name", models.CharField(max_length=25)),
                ("visit_code", models.CharField(max_length=25)),
                ("model", models.CharField(max_length=50)),
                ("current_entry_title", models.CharField(max_length=250, null=True)),
                ("show_order", models.IntegerField()),
                (
                    "entry_status",
                    models.CharField(
                        choices=[
                            ("REQUIRED", "New"),
                            ("KEYED", "Keyed"),
                            ("MISSED", "Missed"),
                            ("NOT_REQUIRED", "Not required"),
                        ],
                        db_index=True,
                        default="REQUIRED",
                        max_length=25,
                    ),
                ),
                ("due_datetime", models.DateTimeField(blank=True, null=True)),
                ("report_datetime", models.DateTimeField(blank=True, null=True)),
                (
                    "entry_comment",
                    models.TextField(blank=True, max_length=250, null=True),
                ),
                ("close_datetime", models.DateTimeField(blank=True, null=True)),
                ("fill_datetime", models.DateTimeField(blank=True, null=True)),
                ("panel_name", models.CharField(max_length=50, null=True)),
            ],
            options={
                "verbose_name": "Requisition Metadata",
                "verbose_name_plural": "Requisition Metadata",
                "abstract": False,
            },
        ),
        migrations.AlterUniqueTogether(
            name="requisitionmetadata",
            unique_together=set(
                [
                    (
                        "subject_identifier",
                        "visit_schedule_name",
                        "schedule_name",
                        "visit_code",
                        "model",
                        "panel_name",
                    )
                ]
            ),
        ),
        migrations.AlterUniqueTogether(
            name="crfmetadata",
            unique_together=set(
                [
                    (
                        "subject_identifier",
                        "visit_schedule_name",
                        "schedule_name",
                        "visit_code",
                        "model",
                    )
                ]
            ),
        ),
    ]



# Generated by Django 2.0.7 on 2018-07-06 13:19

from django.db import migrations


class Migration(migrations.Migration):
    dependencies = [("edc_metadata", "0009_auto_20180116_1528")]

    operations = [
        migrations.AlterModelOptions(
            name="crfmetadata",
            options={
                "verbose_name": "Crf Metadata",
                "verbose_name_plural": "Crf Metadata",
            },
        ),
        migrations.AlterModelOptions(
            name="requisitionmetadata",
            options={
                "verbose_name": "Requisition Metadata",
                "verbose_name_plural": "Requisition Metadata",
            },
        ),
    ]



from __future__ import annotations

from typing import TYPE_CHECKING, Type

from django.apps import apps as django_apps
from django.conf import settings

from .constants import CRF, REQUISITION

if TYPE_CHECKING:
    from edc_metadata.models import CrfMetadata, RequisitionMetadata


def get_crf_metadata_model_cls() -> Type[CrfMetadata]:
    return django_apps.get_model("edc_metadata.crfmetadata")


def get_requisition_metadata_model_cls() -> Type[RequisitionMetadata]:
    return django_apps.get_model("edc_metadata.requisitionmetadata")


def get_metadata_model_cls(
    metadata_category: str,
) -> Type[CrfMetadata] | Type[RequisitionMetadata]:
    if metadata_category == CRF:
        model_cls = get_crf_metadata_model_cls()
    elif metadata_category == REQUISITION:
        model_cls = get_requisition_metadata_model_cls()
    else:
        raise ValueError(f"Invalid metadata category. Got {metadata_category}.")
    return model_cls


def verify_model_cls_registered_with_admin():
    return getattr(settings, "EDC_METADATA_VERIFY_MODELS_REGISTERED_WITH_ADMIN", False)


def refresh_references_and_metadata_for_timepoint(
    appointment_or_related_visit, skip_references: bool | None = None
):
    """Refresh references and metadata for the given timepoint.

    Note: unlike the signal we call `update_references_on_save`
    (note the `s` in `references`) instead.

    See also `metadata_create_on_post_save`, `ReferenceModelMixin`
    and `CreatesMetadataModelMixin`.
    """
    if appointment_or_related_visit:
        try:
            related_visit = appointment_or_related_visit.related_visit
        except AttributeError:
            related_visit = appointment_or_related_visit
        if not skip_references:
            related_visit.update_references_on_save()
        related_visit.metadata_create()
        if django_apps.get_app_config("edc_metadata").metadata_rules_enabled:
            related_visit.run_metadata_rules()



from typing import Tuple

from django.contrib import admin

from ..admin_site import edc_metadata_admin
from ..models import RequisitionMetadata
from .modeladmin_mixins import MetadataModelAdminMixin


@admin.register(RequisitionMetadata, site=edc_metadata_admin)
class RequisitionMetadataAdmin(MetadataModelAdminMixin):
    change_list_title = "Requisition collection status"
    change_form_title = "Requisition collection status"

    @staticmethod
    def panel(obj=None):
        return obj.panel_name

    def get_search_fields(self, request):
        search_fields = list(super().get_search_fields(request))
        search_fields.append("panel_name")
        return tuple(search_fields)

    def get_list_display(self, request) -> Tuple[str, ...]:
        list_display = list(super().get_list_display(request))
        list_display.insert(3, "panel_name")
        return tuple(list_display)

    def get_list_filter(self, request) -> Tuple[str, ...]:
        list_filter = list(super().get_list_filter(request))
        list_filter.insert(1, "panel_name")
        return tuple(list_filter)



from __future__ import annotations

from typing import TYPE_CHECKING, Any
from warnings import warn

from django.apps import apps as django_apps
from django.conf import settings
from django.core.exceptions import ObjectDoesNotExist

from ..constants import DO_NOTHING
from .predicate import NoValueError

if TYPE_CHECKING:
    from edc_registration.models import RegisteredSubject
    from edc_visit_tracking.model_mixins import VisitModelMixin as Base

    from ..model_mixins.creates import CreatesMetadataModelMixin
    from .logic import Logic

    class RelatedVisitModel(CreatesMetadataModelMixin, Base):
        pass


class RuleEvaluatorError(Exception):
    pass


class RuleEvaluatorRegisterSubjectError(Exception):
    pass


show_edc_metadata_warnings = getattr(settings, "EDC_METADATA_SHOW_NOVALUEERROR_WARNING", False)


class RuleEvaluator:

    """A class to evaluate a rule.

    Sets `self.result` to REQUIRED, NOT_REQUIRED or None.

    Set as a class attribute on Rule.

    Ensure the `model.field` is registered with `site_reference_configs`.

    Note: the predicate, which is a callable, will create a Reference
    model instance if it does not exist. See `ReferenceGetter` in
    `edc_reference`.

    See also RuleGroup and its metaclass.
    """

    def __init__(
        self, logic: Logic = None, related_visit: RelatedVisitModel = None, **kwargs
    ) -> None:
        self._registered_subject: RegisteredSubject | None = None
        self.logic: Logic = logic
        self.result: str | None = None
        self.related_visit = related_visit
        options = dict(
            visit=self.related_visit, registered_subject=self.registered_subject, **kwargs
        )
        try:
            predicate = self.logic.predicate(**options)
        except NoValueError as e:
            if show_edc_metadata_warnings:
                warn(
                    f"{str(e)} To ignore set settings."
                    "EDC_METADATA_SHOW_NOVALUEERROR_WARNING=False."
                )
            pass
        else:
            if predicate:
                if self.logic.consequence != DO_NOTHING:
                    self.result = self.logic.consequence
            else:
                if self.logic.alternative != DO_NOTHING:
                    self.result = self.logic.alternative

    @property
    def registered_subject_model(self) -> Any:
        app_config = django_apps.get_app_config("edc_registration")
        return app_config.model

    @property
    def registered_subject(self) -> Any:
        """Returns a registered subject model instance or raises."""
        if not self._registered_subject:
            try:
                self._registered_subject = self.registered_subject_model.objects.get(
                    subject_identifier=self.related_visit.subject_identifier
                )
            except ObjectDoesNotExist as e:
                raise RuleEvaluatorRegisterSubjectError(
                    f"Registered subject required for rule {repr(self)}. "
                    f"subject_identifier='{self.related_visit.subject_identifier}'. "
                    f"Got {e}."
                )
        return self._registered_subject
